#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
import ConfigParser
import StringIO
import Queue
from driver.hfcrypto import HfCrypto

_logger = logging.getLogger(__name__)
ENABLE_DEBUG_LOG = True


class HeatFileManager(object):
    """!
    Management der Heizdateien.
    Kann Dateien lesen und in Dictionaries umwandeln.
    Kann Dictionaries erstellen und in Dateien speichern.
    """

    def __init__(self, config_manager):
        """!
        Initialisiert den Manager für Heizdateien.
        @param config_manager: Instanz der Klasse ConfigManager
        """

        _logger.debug("HeatFileManager init")
        self._name = "HeatFileManager"

        self._output_buffer = Queue.Queue()

        self._hfStages = []

        self._material = ''             # Gewünschtes Material
        self._prev_material = ''        # Zuletzt verwendetes Material
        self._material_changed = False  # True, wenn Materialwechsel erkannt

        self._max_heatingrate = int(config_manager.get_config_file_value('heater', 'max_heatingrate', '99'))
        self._max_temp = int(config_manager.get_config_file_value('temp_limits', 'limit-temperature-cc', '1953'))
        self._max_temp -= 30
        self._hf_directory = config_manager.get_parser().get('path', 'hf_directory')

    def safeHeatData(self, hf_data):
        """!
        Speichert die übergebenen Daten in eine Heizdatei ab. Der Name der Datei steht in den Metainfos.
        @param hf_data: Heizprofil, das in eine Datei gespeichert werden soll.
        @return True - Heizprofil erfolgreich gespeichert.\n
                False -  Heizprofil konnte nicht gespeichert werden.
        """
        retval = False

        hf_data = hf_data.replace(';', '\n')
        buf = StringIO.StringIO(hf_data)                                # Pseudo Datei mit Heizdaten erzeugen
        heatfileconfig = ConfigParser.ConfigParser()                    # ConfigParser für ini-Dateien
        heatfileconfig.readfp(buf)                                      # Pseudo Datei in ConfigParser laden

        if heatfileconfig.has_section('metainfo'):                      # Datei benötigt metainfo
            if ENABLE_DEBUG_LOG:
                _logger.debug('metainfo: ' + str(heatfileconfig.items('metainfo')))
            hf_name = heatfileconfig.get('metainfo', 'name') + '.hf'    # Dateiname zusammenbauen
            hf_path = self._hf_directory + 'user/' + hf_name            # Dateipfad zusammenbauen
            try:
                hf_file = open(hf_path, "w")                            # Datei zum schreiben öffnen
            except IOError:
                _logger.error("error: cannot open file " + hf_path)
            else:
                hf_file.write(hf_data)                                  # in Datei schreiben
                hf_file.close()                                         # Anschließend Datei schließen
                retval = True
        else:
            _logger.error('heat data not valid, metainfo is missing!')

        return retval

    def loadHeatData(self, hf_name):
        """!
        Lädt den Inhalt einer Heizdatei und gibt diesen als String zurück.
        @param hf_name: Name der Heizdatei.
        @return Inhalt der Heizdatei. Leer, wenn Datei ungültig oder nicht vorhanden.
        """
        retval = ''

        hf_path = self._getHeatFilePath(hf_name)                      # Pfad zur Heizdatei suchen
        if hf_path == '':                                             # Datei nicht gefunden
            return ''                                                 # ohne Ergebnis beenden

        try:
            hf_file = open(hf_path, "r")                              # Datei zum lesen öffnen
        except IOError:
            _logger.error("error: cannot open file " + hf_path)
        else:
            _logger.debug("opened file " + hf_path)
            hf_lines = hf_file.readlines()  # Inhalt in eine Liste lesen
            hf_file.close()  # Anschließend Datei schließen

            heatfileconfig = ConfigParser.ConfigParser()  # ConfigParser für ini-Dateien
            heatfileconfig.read(hf_path)  # angegebene Datei lesen

            if heatfileconfig.has_option('metainfo', 'encrypted'):    # Prüfen, ob Datei verschlüsselt ist
                hfcrypto = HfCrypto()                                 # Falls ja,
                heatfileconfig = hfcrypto.decrypt(heatfileconfig)     # dann entschlüsseln
                decryptedHeatFile = StringIO.StringIO()               # Pseudo Datei erstellen
                heatfileconfig.write(decryptedHeatFile)               # Entschlüsselte Daten dort rein
                hf_lines = decryptedHeatFile.getvalue()               # und danach Inhalt in eine Liste lesen
                hf_lines = hf_lines.replace(' ', '')                  # Keine Leerzeichen erlaubt!
                hf_lines = hf_lines.replace('\n', ';')                # alle LF durch ; ersetzen
                hf_lines = hf_lines.replace('\r', '')                 # alle CR entfernen
                retval = hf_lines
            else:
              for hf_line in hf_lines:                                  # Zeilenweise
                hf_line = hf_line.split(';')[0]                       # alle Kommentare entfernen
                hf_line = hf_line.replace(' ', '')                    # Keine Leerzeichen erlaubt!
                hf_line = hf_line.replace('\n', ';')                  # alle LF durch ; ersetzen
                hf_line = hf_line.replace('\r', '')                   # alle CR entfernen
                retval += hf_line                                     # dann an Rückgabe hängen

        return retval

    def deleteHeatData(self, hf_name):
        """!
        Löscht die angegebene Heizdatei.
        @param hf_name: Name der Heizdatei.
        @return True, wenn Löschen erfolgreich.
        """
        retval = False

        hf_path = self._hf_directory + 'user/' + hf_name + '.hf'      # Dateipfad zusammenbauen
        try:
            os.remove(hf_path)                                        # Datei löschen
        except IOError:
            _logger.error("error: cannot delete file " + hf_path)
        else:
            _logger.info("deleted file " + hf_path)
            retval = True

        return retval

    def has_material_changed(self):
        """!
        Gibt True zurück, wenn beim Lesen einer Heizdatei erkannt wurde, dass sich das Material geändert hat.
        Nach einem Materialwechsel ist ein Wechsel der Sinterschale notwendig.
        @return True, wenn sich das Material geändert hat.
        """
        return self._material_changed

    def confirm_material_change(self, conf=True):
        """!
        Durch den Aufruf dieser Funktion wird der Materialwechsel und der Wechsel der Sinterschale bestätigt.
        """
        if conf == True:
            self._prev_material = self._material
        else:
            self._material = self._prev_material
        self._material_changed = False

    def read(self):
        """!
        Mit dieser Funktion holt der Gs1Controller Meldungen des Heat File Managements an die GUI ab.
        @return Meldungen des Heat File Managements an die GUI. Wenn None, liegt keine Meldung vor
        """
        if self._output_buffer.empty():
            return None
        else:
            rc = self._output_buffer.get()
            self._output_buffer.task_done()
            return rc

    def getMaxHeatRate(self):
        """!
        Gibt den Wert der maximal erlaubten Heizrate zurück.
        @return Maximal erlaubte Heizrate in K/min.
        """
        return self._max_heatingrate

    def hfToFavorite(self, hf_name, favorite):
        """!
        Markiert eine Heizdatei als Favorit oder nicht.
        @param hf_name: Name der Heizdatei.
        @param favorite: True - Datei wird als Favorit markiert\n
                         False - Datei wird nicht als Favorit markiert.
        @return True, wenn Favorit erfolgreich gesetzt oder gelöscht.
        """
        hf_path = self._getHeatFilePath(hf_name)            # Pfad zur Heizdatei suchen
        if hf_path == '':                                   # Datei nicht gefunden
            return False                                    # ohne Ergebnis beenden

        heatfileconfig = ConfigParser.ConfigParser()        # ConfigParser für ini-Dateien
        heatfileconfig.read(hf_path)                        # angegebene Datei lesen

        if heatfileconfig.has_section('metainfo'):          # Datei benötigt metainfo
            heatfileconfig.set('metainfo', 'favorite', favorite)
            with open(hf_path, 'wb') as heatFile:
                heatfileconfig.write(heatFile)
            return True                                     # metainfo geschrieben, alles OK
        else:
            _logger.error('Heatfile not valid, metainfo is missing!')
            return False                                    # ohne Ergebnis beenden

    def listAllHeatFileNames(self, user_only=False):
        """!
        Lädt die Namen aller Heizdateien und gibt diese als Liste zurück.
        @return Liste aller vorhandenen Heizprofile.
        """
        hf_list = []
        for (dirpath, dirnames, filenames) in os.walk(self._hf_directory):
            for file_ in filenames:
                hf_name = os.path.splitext(file_)[0]
                if user_only:
                    if dirpath.endswith('user'):
                        hf_list.append(hf_name)
                else:
                    hf_list.append(hf_name)

        hf_list.sort(reverse=True)
        _logger.debug(hf_list)

        return hf_list

    def _getHeatFilePath(self, hf_name):
        """!
        Sucht die angegebene Heizdatei und gibt den Pfad des ersten Treffers zurück.
        @param hf_name: Heizdateiname für die der Dateipfad gesucht werden soll.
        @return Pfad zur angegebenen Heizdatei.
        """
        hf_name = hf_name + '.hf'                     # Dateiname zusammenbauen

        for (dirpath, dirnames, filenames) in os.walk(self._hf_directory):
            for file_ in filenames:
                if file_ == hf_name:
                    return os.path.join(dirpath, file_)

        _logger.error("file {} not found".format(hf_name))

        return ''       # Schleife wurde ohne Treffer beendet, leeren Pfad ausgeben

    def parseHeatFile(self, hf_name):
        """!
        Liest die Daten aus einer Heizdatei.
        @param hf_name: Name der Heizdatei.
        @return Liste mit dem Heizprofil der angegebenen Datei. Ist die Datei fehlerhaft, wird eine leere Liste zurückgegeben.
        """
        self._hfStages = []                                 # Neue Datei wird geöffnet, daher alte Liste leeren

        hf_path = self._getHeatFilePath(hf_name)            # Pfad zur Heizdatei suchen
        if hf_path == '':                                   # Datei nicht gefunden
            return []                                       # ohne Ergebnis beenden

        heatfileconfig = ConfigParser.ConfigParser()        # ConfigParser für ini-Dateien
        heatfileconfig.read(hf_path)                        # angegebene Datei lesen

        if heatfileconfig.has_section('metainfo'):          # Datei benötigt metainfo
            if ENABLE_DEBUG_LOG:
                _logger.debug('metainfo: ' + str(heatfileconfig.items('metainfo')))

            self._material = heatfileconfig.get('metainfo', 'material')   # Materialinfo aus metadaten lesen

            if heatfileconfig.has_option('metainfo', 'encrypted'):
                hfcrypto = HfCrypto()
                heatfileconfig = hfcrypto.decrypt(heatfileconfig)

            heatfileconfig.remove_section('metainfo')       # metainfo wurde gelesen, daher nicht mehr benötigt
        else:
            _logger.error('Heatfile not valid, metainfo is missing!')
            return []

        hf_sections = heatfileconfig.sections()             # Alle Stufen erfassen
        hf_sections.sort()                                  # und sortieren

        # Info: Doppelte Stufen werden in der Reihenfolge durchlaufen, die sie in der Datei haben.

        for heatStage in hf_sections:                       # alle verbleibenden Sections durchlaufen
            stage = {}                                      # leeren
            try:
                int(heatStage)                              # nur Zahlen erlaubt
            except ValueError:
                _logger.warning('Invalid heatstage: ' + str(heatStage))
                continue

            heatstageparams = heatfileconfig.options(heatStage)     # verfügbare Parameter lesen
            for parameter in heatstageparams:                       # alle gefundenen Parameter verwendet
                if parameter == 'temp':                 # Temperatur
                    try:
                        temp = int(heatfileconfig.get(heatStage, parameter))    # als Integer sichern
                        if temp in range(273, self._max_temp+1):    # erlaubter Bereich 273K..self._max_temp
                            stage[parameter] = temp
                    except ValueError:
                        pass                            # bei Fehler nichts machen, Stufe wird weiter unten aussortiert
                elif parameter == 'rate':               # Heiz-/Kühlrate
                    try:
                        rate = int(heatfileconfig.get(heatStage, parameter))    # als Integer sichern
                        rate = min(rate, self._max_heatingrate)            # Begrenzen auf Maximalwert aus Config File
                        if rate > 0:                    # muss nach Begrenzung größer 0 sein
                            stage[parameter] = rate
                    except ValueError:
                        pass                            # bei Fehler nichts machen, Stufe wird weiter unten aussortiert
                elif parameter == 'hold':               # Haltezeit
                    try:
                        hold = int(heatfileconfig.get(heatStage, parameter))    # als Integer sichern
                        if hold >= -1:                  # muss >= -1 sein
                            stage[parameter] = hold
                    except ValueError:
                        pass                            # bei Fehler nichts machen, Stufe wird weiter unten aussortiert
                elif parameter in ['linearcooling', 'argon', 'vacuum', 'forcecooling', 'vent', 'res1', 'res2', 'res3', 'res4']:     # Alle anderen
                    stage[parameter] = True if heatfileconfig.get(heatStage, parameter) == 'on' else False                              # als Boolean, Standard ist False
                elif parameter == 'door':
                    doorstate = heatfileconfig.get(heatStage, parameter)
                    if doorstate in ['open', 'close', 'force_open']:    # Die Türe darf nur open, close oder force_open sein
                        stage[parameter] = heatfileconfig.get(heatStage, parameter)
                    else:                                # falls nicht, immer zu machen (close)
                        stage[parameter] = 'close'
                else:                                    # Ungültige Parameter führen zu einem Logeintrag
                    _logger.warning('Invalid parameter: ' + str(parameter))

            if ENABLE_DEBUG_LOG:
                _logger.debug(stage)
            if ('temp' in stage) and ('rate' in stage) and ('hold' in stage):   # Es sind mindestens die Parameter temp, rate und hold gefordert
                self._hfStages.append(stage)            # alle Parameter der aktuellen Stufe an Liste anhängen
            else:
                _logger.warning('Invalid heatstage {} - parameter set incomplete: {}'.format(heatStage, stage))

        if self._hfStages:
            # Bei Materialwechsel eine Meldung an GUI veranlassen
            if self._prev_material != self._material:
                self._material_changed = True
                self._output_buffer.put("matchanged=={}".format(self._material.lower()))

        return self._hfStages
